#!/usr/bin/env python
import re

# build a simple binomial table
def build_cnk(l):
    c = [[0 for j in xrange(i+1)] for i in xrange(l+1)]
    for i in xrange(l+1):
        c[i][0] = 1;
        c[i][i] = 1;
        for j in xrange(1,i):
            c[i][j] = c[i-1][j-1] + c[i-1][j];
    return c

# simplify a fraction
def frac_gcd(num, den):
    a = num
    b = den
    while b:
        a, b = b, a%b    
    return (num//a,den//a)


# sorting functions (from http://stackoverflow.com/questions/5967500/how-to-correctly-sort-a-string-with-a-number-inside)
def atoi(text):
    return int(text) if text.isdigit() else text

def natural_keys(text):
    '''
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    '''
    return [ atoi(c) for c in re.split('(\d+)', text) ]

# seed nomalizer : takes the smaller alphabetic order
def normalize(s):
    s_r = s[::-1]
    ls   = s.split(',');   ls.sort(key=natural_keys);   s   = ','.join(ls)  
    ls_r = s_r.split(','); ls_r.sort(key=natural_keys); s_r = ','.join(ls_r)
    if s < s_r:
        return s
    else:
        return s_r

# print seeds
def print_seedlist(sl,fo,frac):
    fo.write('[')
    for i,e in enumerate(sl):
        fo.write("'"+(str(e).replace(',','_')) +"'")
        if i < len(sl)-1:
            fo.write(",")
    fo.write('] '+str(frac)+"\n")    


# parse a file and extract the pareto according to the discrete condition (0:DIRAC, 1:HEAVISIDE, 2:REVERSE HEAVISIDE)
def parse_ied_file_condition(filename_in,filename_out,l,condition,c):
    '''
    '''
    already_seen = set();

    if condition == 0:
        vals  = [0 for i in xrange(l+1)]
    else:
        vals  = [(0,1) for i in xrange(l+1)]
    seeds = {i:[] for i in xrange(l+1)}
    
    with open(filename_in,"r") as f:
        for line in f:
            m = re.search('^([01,]*)\s+[0-9.]+\s+[0-9.]+\s+[0-9.]+\s+(.*)$',line)
            seed = str(m.group(1))
            seed = normalize(seed)
            if seed in already_seen:
                continue
            else:
                already_seen.add(seed)
                
            p = re.compile('([0-9]+,[0-9]+=[0-9]+);')
            coefs = p.findall(str(m.group(2)))

            if condition == 0:
                seed_vals = [0 for i in xrange(l+1)]
            else:
                seed_vals = [(0,1) for i in xrange(l+1)]

            first_coef = True
            for coef in coefs:
                m2 = re.search('^([0-9]+),1=([0-9]+)$',coef)
                if m2:
                    index = int(m2.group(1))
                    count = int(m2.group(2))

                    # (0) DIRAC (quite simple, direct evaluation without postprocessing)
                    if condition == 0:
                        #>> solve the case when small values are not occuring on the dirac function
                        if first_coef:
                            for u in xrange(index):
                                seeds[u].append(seed)
                                seeds[u].sort(key=natural_keys)
                            first_coef = False
                        #<<
                        if vals[index] == count :
                            seeds[index].append(seed)
                            seeds[index].sort(key=natural_keys)
                        elif vals[index] < count:
                            seeds[index] = [seed]
                            vals[index] = count                    
                    else: # (1) HEAVISIDE or (2) REVERSE HEAVISIDE (first preprocess and keep in a list)
                        seed_vals[index] = frac_gcd(count,c[l][index])
                        
            # HEAVISIDE (1) : (postprocess by sum)
            if condition == 1:
                frac_val = (0,1)
                for i in xrange(l+1):
                    frac_val = frac_gcd( frac_val[0] * seed_vals[i][1] + seed_vals[i][0] * frac_val[1],  frac_val[1] * seed_vals[i][1])
                    if frac_val[0] * vals[i][1] == frac_val[1] * vals[i][0]:
                        seeds[i].append(seed)
                        seeds[i].sort(key=natural_keys)
                    elif frac_val[0] * vals[i][1] > frac_val[1] * vals[i][0]:
                            seeds[i] = [seed]
                            vals[i] = frac_val
                            
            # REVERSE HEAVISIDE (2) : (postprocess by reverse-sum)
            elif condition == 2:
                frac_val = (0,1)
                for i in xrange(l,-1,-1):
                    frac_val = frac_gcd( frac_val[0] * seed_vals[i][1] + seed_vals[i][0] * frac_val[1],  frac_val[1] * seed_vals[i][1])
                    if frac_val[0] * vals[i][1] == frac_val[1] * vals[i][0]:
                        seeds[i].append(seed)
                        seeds[i].sort(key=natural_keys)
                    elif frac_val[0] * vals[i][1] > frac_val[1] * vals[i][0]:
                            seeds[i] = [seed]
                            vals[i] = frac_val

    # Sort the seeds before the output
    for i in xrange(l+1):
        seeds[i].sort(key=natural_keys)

    # PARETO OUTPUT
    with open(filename_out, 'w') as fo:
        for i in xrange(l):
            if set(seeds[i]) != set(seeds[i+1]):
                print_seedlist(seeds[i],fo,str(float(i+1)/float(l)))
                # print previous seeds (until value "float(i+1)/float(l)")
        print_seedlist(seeds[l],fo,"1.0")



# MAIN
conds = ["dirac","heaviside","reverse_heaviside"]
for n in xrange(1,5):
    for w in xrange(3,17):
        for l in xrange(2*w,65):
            for c,cn in enumerate(conds):
                print "_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+".ied",  "_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+"_"+cn+"_python.txt"
                CNK = build_cnk(l)
                parse_ied_file_condition("experiments/_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+".ied","experiments/_n"+str(n)+"_w"+str(w)+"_s"+str(2*w)+"_l"+str(l)+"_"+cn+"_python.txt",l,c,CNK)

